package piglets.alpha;

import java.util.Vector;
import java.util.TreeSet;

import universum.bi.*;

/**
 * Created by IntelliJ IDEA.
 * User: Tanya
 * Date: 02.01.2007
 * Time: 21:02:03
 */

class DrawWell implements Comparable<DrawWell> {
    public float            growth;
    public float            lastCount;
    public Location         pos;
    public final TreeSet<Piglet> owner = new TreeSet<Piglet>();

    public DrawWell(float growth, float lastCount, Location pos) {
        this.growth     = growth;
        this.lastCount  = lastCount;
        this.pos        = pos;
    }

    public int compareTo(DrawWell o) {
        if (growth > o.growth) {
            return -1;
        }
        if (growth < o.growth) {
            return 1;
        }
        return 0;
    }
}

class WorldCell {
    public boolean  explored;
    public int      toBeExplored;
    public DrawWell drawWell;
}

public class Manager {
    public static final float S0  = 3.f;
    public static final float M0  = 5.f;
    public static final float R_avial = 3.f;
    public static final float R_emin = 1.2f;

    public static final int MASS_GROW_THRESHOLD = 20;
    public static final int SHARE_THRESHOLD     = 150;

    public static final float INF   = 1e6f;
    public static final float EPS   = 1e-5f;

    private int turnCounter = 0;
    private int turn = 0;

    public WorldCell[][] world;
    public final Vector<Piglet> piglets = new Vector<Piglet> ();
    public final TreeSet<DrawWell> drawWells = new TreeSet<DrawWell> ();

    private void makeTurn(BeingInterface bi) {
        // update exploration info and clear orders
        for (Piglet p : piglets) {
            PointInfo curInfo = bi.getPointInfo(p);
            updateWorld(curInfo, p);

            for (PointInfo info : bi.getNeighbourInfo(p)) {
                updateWorld(info, p);
            }
        }

        if (piglets.size() > SHARE_THRESHOLD) {
            // check if piglet about to die
            for (Piglet p1 : piglets) {
                float energy = bi.getEnergy(p1);
                if (energy < p1.M * Constants.K_emin * R_emin) {
                    float wantsTo = p1.M * Constants.K_emin * R_emin * R_emin - energy;

                    for (Piglet p2 : piglets) {
                        float energy2 = bi.getEnergy(p2);
                        if (p2.order != Order.SHARE && energy2 > p2.M * Constants.K_emin * R_emin * R_emin * R_emin) {
                            p2.order = Order.SHARE;
                            p2.shareTarget = p1;
                            p2.shareAmount = wantsTo;

                            // if it has alpha draw-well
                            if (p2.ownedWell != null) {
                                p2.ownedWell.owner.remove(p2);
                                p2.ownedWell = null;
                            }
                            break;
                        }
                    }
                }
            }
        }

        // try to find closest piglet for free draw-weels
        for (DrawWell dw : drawWells) {
            if (dw.lastCount + dw.growth * 10.f < 5.f && piglets.size() < 20) {
                continue;
            }

            if (!dw.owner.isEmpty()) {
                // check if it's good draw-well
                float toEat = 0;
                float avail = -INF;
                for (Piglet p : dw.owner) {
                    toEat += p.M - bi.getEnergy(p);

                    if (bi.getLocation(p).equals(dw.pos)) {
                        // if the same position, get info
                        PointInfo pi = bi.getPointInfo(p);
                        avail = pi.getCount(p);
                    }
                }

                // if at least one piglet at draw-well
                if (avail < 0.f || avail < R_avial * toEat) {
                    continue;
                }
            }

            float minDistance = INF;
            Piglet minPiglet = null;
            for (Piglet p : piglets) {
                if (p.ownedWell != null) {
                    continue;
                }

                Location pos = bi.getLocation(p);
                float distance = bi.distance(pos, dw.pos);

                if (distance < minDistance) {
                    minPiglet = p;
                    minDistance = distance;
                }
            }

            if (minPiglet != null) {
                minPiglet.order = Order.DRAWWELL;
                minPiglet.ownedWell = dw;
                dw.owner.add(minPiglet);
            } else {
                // nothing to find, all piglets are busy
                break;
            }
        }
    }

    public void init(UserGameInfo info) {
        this.turn = 0;
        this.turnCounter = 0;
        this.piglets.clear();
        this.drawWells.clear();

        world = new WorldCell[Constants.getWidth()][Constants.getHeight()];

        for (int i=0; i<world.length; ++i) {
            for (int j=0; j<world[i].length; ++j) {
                world[i][j] = new WorldCell();
            }
        }
    }

    public synchronized void updateTurn(BeingInterface bi, Piglet pig) {
        if (turnCounter++ == 0) {
            ++turn;
            makeTurn(bi);
        }
        if (turnCounter == piglets.size()) {
            turnCounter = 0;
        }
    }

    public synchronized int getTurn() {
        return turn;
    }

    public synchronized void register(Piglet p) {
        piglets.add(p);
    }

    public synchronized void unregister(Piglet p) {
        if (p.ownedWell != null) {
            p.ownedWell.owner.remove(p);
            --turnCounter;
        }
        piglets.remove(p);
    }

    private void updateWorld(PointInfo info, Piglet p) {
        Location pos    = info.getLocation();
        WorldCell cell  = world[pos.getX()][pos.getY()];

        if (cell.explored) {
            return ;
        }
        cell.explored   = true;
        float growth    = info.getGrowthRate(p);

        if (growth > 0.f) {
            cell.drawWell = new DrawWell(growth, info.getCount(p), pos);
            drawWells.add(cell.drawWell);
        }
    }
}
