package piglets.alpha;

import universum.bi.*;
import universum.util.Util;

/**
 * Created by IntelliJ IDEA.
 * User: Tanya
 * Date: 02.01.2007
 * Time: 21:02:15
 */

enum Order {
    EXPLORE,
    DRAWWELL, SHARE,
}

public class Piglet implements Being, Comparable<Piglet> {
    private static final Manager manager = new Manager();

    public Integer id;
    public float M;
    public float S;

    public DrawWell ownedWell;
    public Order order;
    public Piglet shareTarget;
    public float shareAmount;

    public void reinit(UserGameInfo userGameInfo) {
        manager.init(userGameInfo);
    }

    public String getName() {
        return "Piglet";
    }

    public String getOwnerName() {
        return "E13";
    }

    public BeingParams getParams() {
        if (manager.piglets.size() >= Manager.MASS_GROW_THRESHOLD) {
            return new BeingParams(M * Constants.K_maxbornvariation, Manager.S0);
        }
        return new BeingParams(Manager.M0, Manager.S0);
    }

    public void processEvent(BeingInterface bi, Event event) {
        switch (event.kind()) {
            case BEING_BORN:
                manager.register(this);
                BeingParams bp = (BeingParams)event.param();
                M = bp.M;
                S = bp.S;
                id = bi.getId(this);
                break;
            case BEING_DEAD:
                manager.unregister(this);
                break;
        }
    }

    public Event makeTurn(BeingInterface bi) {
        manager.updateTurn(bi, this);

        Location curPos = bi.getLocation(this);

        if (order == Order.SHARE) {
            Event e = new Event(EventKind.ACTION_GIVE, (int)shareAmount, shareTarget);
            shareTarget = null;
            order = Order.EXPLORE;
            return e;
        }

        if (order == Order.DRAWWELL) {
            Location next = findNextOnPath(curPos, ownedWell.pos, bi);
            if (next != null) {
                return new Event(EventKind.ACTION_MOVE_TO, next);
            }

            // we already at draw-well, let's eat or born
            if (bi.getEnergy(this) > M * Constants.K_toborn) {
                return new Event(EventKind.ACTION_BORN, getParams());
            }

            return new Event(EventKind.ACTION_EAT, M);
        }

        // try to find turn to explore
        Location nextUnexplored = findUnexploredTurn(curPos);
        if (nextUnexplored != null) {
            return new Event(EventKind.ACTION_MOVE_TO, nextUnexplored);
        }

        return null;
    }

    private Location findNextOnPath(Location curPos, Location pos, BeingInterface bi) {
        float dist = Math.abs(bi.distance(curPos, pos));

        if (dist < Manager.EPS) {
            return null;
        }

        if (dist < S + Manager.EPS) {
            return pos;
        }

        final float [] dx = new float[] { 0, 1, 0, 1, 0, -1, -1 };
        final float [] dy = new float[] { 0, 0, 1, 1, -1, 0, -1 };

        float x0 = curPos.getX();
        float y0 = curPos.getY();
        float x1 = pos.getX();
        float y1 = pos.getY();

        int minIndex = -1;
        float minDistance = Manager.INF;

        for (int i=0; i<dx.length; ++i) {
            float x = x1 + Constants.getWidth()  * dx[i];
            float y = y1 + Constants.getHeight() * dy[i];

            dist = Math.max(Math.abs(x - x0), Math.abs(y - y0));
            if (dist < minDistance) {
                minDistance = dist;
                minIndex = i;
            }
        }

        assert minIndex != -1;
        x1 += Constants.getWidth()  * dx[minIndex];
        y1 += Constants.getHeight() * dy[minIndex];

        float x_ar = x1 - x0;
        float y_ar = y1 - y0;

        if (Math.abs(x_ar) > S) {
            x_ar = S * Math.signum(x_ar);
        }
        if (Math.abs(y_ar) > S) {
            y_ar = S * Math.signum(y_ar);
        }

        return new Location(curPos.getX() + (int)x_ar, curPos.getY() + (int)y_ar);
    }

    public int compareTo(Piglet o) {
        return id.compareTo(o.id);
    }

    private class NextTurnInfo {
        public int nextLeft, nextTop;
        public int left, top;
        public int unexplored;

        public NextTurnInfo(int nextLeft, int nextTop, int left, int top) {
            this.nextLeft = nextLeft;
            this.nextTop = nextTop;
            this.left = left;
            this.top = top;
            this.unexplored = 0;
        }
    }

    private Location findUnexploredTurn(Location curPos) {
        Piglet.NextTurnInfo[] infoList = new Piglet.NextTurnInfo[] {
            new Piglet.NextTurnInfo( -3,  -3,  -7, -10),
            new Piglet.NextTurnInfo(  0,  -3,  -4, -10),
            new Piglet.NextTurnInfo(  3,  -3,  -1, -10),
            new Piglet.NextTurnInfo(  3,   0,   2,  -4),
            new Piglet.NextTurnInfo(  3,   3,   2,   2),
            new Piglet.NextTurnInfo(  0,   3,  -4,   2),
            new Piglet.NextTurnInfo( -3,   3, -10,   2),
            new Piglet.NextTurnInfo( -3,   0, -10,  -4),
        };

        Piglet.NextTurnInfo maxTurn = null;
        for (Piglet.NextTurnInfo ti : infoList) {
            ti.unexplored = getUnexploredCellsCount(
                curPos.getX() + ti.left, curPos.getY() + ti.top, 9, 9);

            if (maxTurn == null) {
                maxTurn = ti;
            } else {
                if (maxTurn.unexplored < ti.unexplored ||
                maxTurn.unexplored == ti.unexplored && Util.frnd() > 0.9f) {
                    maxTurn = ti;
                }
            }
        }

        assert maxTurn != null;

        if (maxTurn.unexplored == 0) {
            // everything had been explored
            return null;
        }

        return new Location(
                (curPos.getX() + maxTurn.nextLeft + Constants.getWidth()) % Constants.getWidth(),
                (curPos.getY() + maxTurn.nextTop  + Constants.getHeight()) % Constants.getHeight());
    }

    private int getUnexploredCellsCount(int x, int y, int dx, int dy) {
        int count = 0;

        for (int i=0; i<dx; ++i) {
            for (int j=0; j<dy; ++j) {
                int x0 = (x + i + Constants.getWidth()) % Constants.getWidth();
                int y0 = (y + j + Constants.getHeight()) % Constants.getHeight();
                WorldCell cell = manager.world[x0][y0];
                if (!cell.explored && cell.toBeExplored != manager.getTurn()) {
                    ++count;
                }
            }
        }

        return count;
    }
}
