package raptor;

import universum.bi.*;
import universum.util.Util;

/**
 * This is creature for Electric Jungles.
 *
 * @author E13
 */
public class Raptor  implements Being {
    static final Manager manager = new Manager();

    static final float M0   = 5.f;
    static final float S0   = 3.f;
    static final int CREATURE_THRESHOLD = 20;
    private float M;
    private float S;

    // temporaryy variable, just for speed optimization
    private int failCounter = 0;

    // this one is the essence of your being - it's invoked once alpha turn
    // and lets your being do something
    public Event makeTurn(BeingInterface bi) {
        // calc current turn
        synchronized(manager) {
            int turnCounter = manager.getTurnCounter();
            if (turnCounter == 0) {
                manager.setTurn(manager.getTurn() + 1);

                manager.test(bi);
            }
            manager.setTurnCounter(
                    turnCounter + 1 == manager.getCreatureCount() ? 0 : turnCounter + 1);
        }

        // update glade map by known cells
        PointInfo curInfo = bi.getPointInfo(this);
        manager.updateWorld(curInfo, this);

        for (PointInfo pointInfo : bi.getNeighbourInfo(this)) {
            manager.updateWorld(pointInfo, this);

            // let's go to this cell
            if (pointInfo.getGrowthRate(this) > 0.f) {
                synchronized (manager) {
                    Location pos = pointInfo.getLocation();
                    WorldCell cell = manager.world[pos.getX()][pos.getY()];
                    if (cell.owner == null) {
                        cell.owner = this;
                        return new Event(EventKind.ACTION_MOVE_TO, pos);
                    }
                }
            }
        }

        // collect info about this and naighbour cells
        Location curPos = bi.getLocation(this);

        // check if we can born
        final float energy = bi.getEnergy(this);
        if (energy > Constants.K_minbornvariation * M) {
            return new Event(EventKind.ACTION_BORN, getParams());
        }

        // check if we can eat
        final float canEat = M - energy; 
        if (curInfo.getCount(this) > 0) {
            synchronized (manager) {
                WorldCell cell = manager.world[curPos.getX()][curPos.getY()];
                if (cell.growth > 0.f && (cell.owner == null || cell.owner == this)) {
                    cell.owner = this;
                    return new Event(EventKind.ACTION_EAT, canEat);
                }
            }
        }

        // try to explore 
        Location next = findNextTurn(curPos);

        if (next != null) {
            manager.world[next.getX()][next.getY()].toBeExplored = manager.getTurn();
            return new Event(EventKind.ACTION_MOVE_TO, next);
        }

        if (failCounter > 10) {
            return null;
        }

        WorldCell closestCell = null;
        Location  closestLocation = null;
        float minDist = Constants.getWidth() * Constants.getHeight();
        for (int i=0; i<Constants.getWidth(); ++i) {
            for (int j=0; j<Constants.getHeight(); ++j) {
                Location location = new Location(i, j);
                WorldCell cell = manager.world[i][j];

                float distance = Math.abs(bi.distance(curPos, location));
                if (cell.growth > 0.f && cell.owner == null && distance < minDist) {
                    minDist = distance;
                    closestCell = cell;
                    closestLocation = location;
                }
            }
        }

        if (closestCell == null) {
            ++failCounter;
            // nothing found
            return null;
        }

        failCounter = 0;

        if (minDist > S) {
            float dx = closestLocation.getX() - curPos.getX();
            if (Math.abs(dx) > S) {
                dx = Math.signum(dx) * S;
            }
            float dy = closestLocation.getY() - curPos.getY();
            if (Math.abs(dy) > S) {
                dy = Math.signum(dy) * S;
            }
            return new Event(EventKind.ACTION_MOVE_TO, new Location((int)(curPos.getX() + dx), (int)(curPos.getY() + dy)));
        }

        return new Event(EventKind.ACTION_MOVE_TO, closestLocation);
    }

    private class NextTurnInfo {
        public int nextLeft, nextTop;
        public int left, top;
        public int unknownCells;

        public NextTurnInfo(int nextLeft, int nextTop, int left, int top) {
            this.nextLeft = nextLeft;
            this.nextTop = nextTop;
            this.left = left;
            this.top = top;
            this.unknownCells   = 0;
        }
    }

    private Location findNextTurn(Location curPos) {
        NextTurnInfo [] infoList = new NextTurnInfo [] {
            new NextTurnInfo( -3,  -3,  -7, -10),
            new NextTurnInfo(  0,  -3,  -4, -10),
            new NextTurnInfo(  3,  -3,  -1, -10),
            new NextTurnInfo(  3,   0,   2,  -4),
            new NextTurnInfo(  3,   3,   2,   2),
            new NextTurnInfo(  0,   3,  -4,   2),
            new NextTurnInfo( -3,   3, -10,   2),
            new NextTurnInfo( -3,   0, -10,  -4),
        };

        NextTurnInfo maxTurn = null;
        for (NextTurnInfo ti : infoList) {
            ti.unknownCells = calcUnknownCellCount(
                curPos.getX() + ti.left, curPos.getY() + ti.top, 9, 9);

            if (maxTurn == null) {
                maxTurn = ti;
            } else {
                if (maxTurn.unknownCells < ti.unknownCells ||
                maxTurn.unknownCells == ti.unknownCells && Util.frnd() > 0.5f) {
                    maxTurn = ti;
                }
            }
        }

        assert maxTurn != null;

        if (maxTurn.unknownCells == 0) {
            // nothing to do
            return null;
        }

        return new Location(
                (curPos.getX() + maxTurn.nextLeft + Constants.getWidth()) % Constants.getWidth(),
                (curPos.getY() + maxTurn.nextTop  + Constants.getHeight()) % Constants.getHeight());
    }

    private int calcUnknownCellCount(int x, int y, int dx, int dy) {
        int count = 0;

        for (int i=0; i<dx; ++i) {
            for (int j=0; j<dy; ++j) {
                int x0 = (x + i + Constants.getWidth()) % Constants.getWidth();
                int y0 = (y + j + Constants.getHeight()) % Constants.getHeight();
                WorldCell cell = manager.world[x0][y0];
                if (!cell.explored && cell.toBeExplored != manager.getTurn()) {
                    ++count;
                }
            }
        }

        return count;
    }

    public void processEvent(BeingInterface bi, Event e) {
        switch (e.kind()) {
            case BEING_BORN:
                init(bi, bi.getId(this), (BeingParams)e.param());
                break;
            case BEING_DEAD:
                // gonna get dead, decrease population counter
                synchronized (manager) {
                    manager.unregister(this);
                }
                break;
            case BEING_ATTACKED:
                // now just print if one attacks us
                break;
            case BEING_ENERGY_GIVEN:
                // ... or gives alpha gift
                break;
        }
    }

    public String getName() {
        return "Raptor";
    }

    // this method is invoked once per game, to make sure all static are properly inited
    public void reinit(UserGameInfo info) {
        manager.init();
    }

    // return our parameters
    public BeingParams getParams() {
        if (manager.getCreatureCount() > CREATURE_THRESHOLD) {
            return new BeingParams(Math.min(M * Constants.K_maxbornvariation, Constants.K_maxmass), S0); 
        }
        return new BeingParams(M0, S0);
    }

    // who this entity belongs to (and score accounted for)
    public String getOwnerName() {
        return "E13";
    }

    private void init(BeingInterface bi, Integer id, BeingParams bp) {
        // increase population counter
        synchronized(manager) {
            manager.register(this);
        }

        M = bp.M;
        S = bp.S;
    }
}
