package universum.engine;

import java.util.*;
import universum.bi.Location;
import universum.util.*;

/**
 * Default topology of two dimensional torus
 *
 * @author nike
 */
public class DefaultTopology implements Topology {
    private Location[][] space;
    private int width, height, maxDist;
    private Map<Location, List<Location>> locCache;
    private Universe universe;
    private List<Integer> diameters; 
    
    public DefaultTopology(int width, int height) {
        this.width = width;
        this.height = height;
        this.maxDist = Math.max(width, height);
        space = new Location[width][height];
        locCache = new HashMap<Location, List<Location>>();
        diameters = new ArrayList<Integer>();
        diameters.add(width); diameters.add(height); 
	// just dummies, to allow class preloading
	new Sphere(new Location(0, 0), 0);
        new ImmutableList<Location>(new Location[]{});
    }
    
    public void init(Universe u) {
        universe = u;
        for (int i=0; i<width; i++) {
            for (int j=0; j<height; j++) {
                space[i][j] = new Location(i, j);
            }
        }
    }
    
    public List<Location> getNeighbours(Location loc) {
        List<Location> rv = locCache.get(loc);
        if (rv != null) {
            return rv;
        }
        
        synchronized (locCache) {
            int x = loc.getX(); int y = loc.getY();
            int xn = (x - 1 + width) % width; int yn = (y - 1 + height) % height;
            int xp = (x + 1) % width; int yp = (y + 1) % height;
            Location[] locations = new Location[] {
                space[xn][yn], space[x][yn], space[xp][yn],
                space[xn][y],                space[xp][y],
                space[xn][yp], space[x][yp], space[xp][yp]
            };
            rv = new ImmutableList<Location>(locations);
            locCache.put(loc, rv);
        }
        return rv;
    }
    
    public List<Location> getNeighbours(Location loc, float radius) {
        int r = (int)radius;
        if (r == 1) {
            // optimize the common case: nearest neighbour locations are cached
            return getNeighbours(loc);
        }
        return new Sphere(loc, r);
    }
    
    public void walkSpace(Walker<Location> walker) {
        for (int i=0; i<width; i++) {
            for (int j=0; j<height; j++) {
                walker.walk(space[i][j]);
            }
        }
    }

    public Location getRandomLocation() {
        return space[Util.rnd(width)][Util.rnd(height)];
    }
    
    public List<Integer> diameters() {
        return diameters;
    }
     
    public float distance(Location one, Location other) {
        float dx = Math.abs(one.getX() - other.getX());
        dx = Math.min(dx, width - dx);
        float dy = Math.abs(one.getY() - other.getY());
        dy = Math.min(dy, height - dy);
        return Math.max(dx, dy);
    }

    public boolean contains(Location l) {       
        int x = l.getX(), y = l.getY();
        return (x >= 0 && x < width && y >= 0 && y < height);
    }

    private class Sphere extends AbstractList<Location>
        implements List<Location>
    {
        private int x0;
        private int y0;
        private int diameter;
        private int count;

        public Sphere(Location loc, int radius) {
            x0 = loc.getX() - radius;
            y0 = loc.getY() - radius;
            diameter = 2 * radius + 1;
            count = diameter * diameter - 1;
        }

        public int size() {
            return count;
        }

        public Location get(int index) {
            if (index < 0 || index >= count) {
                throw new IndexOutOfBoundsException();
            }
            if (index >= count / 2) {
                // the center of the sphere is omitted
                index++;
            }
            int y = y0 + index / diameter;
            int x = x0 + index % diameter;
            return space[(x + width) % width][(y + height) % height];
        }

    }

}
