package universum.ui;

import java.awt.*;
import java.awt.image.*;
import javax.swing.ImageIcon;

import universum.bi.Being;
import universum.bi.Entity;
import universum.bi.Location;
import universum.bi.Constants;
import universum.engine.*;
import universum.util.*;

/**
 *
 * @author nike
 */
public class Renderer {
    RenderingInfo ri;   
    
    // those are transient fields, valid only during rendering operation
    Graphics2D g;
    // those are cache
    int w, h, x0, y0;
    float sx, sy;
    EntitiesRenderer entitiesRend;
    EnergyRenderer   energyRend;    
    Image beings[][], resources[];
    ImageIcon beingIcon[];
    Color playfieldColor =  new Color(0x9f, 0xb9, 0xff);
        // new Color(0xa1, 0xa1, 0xf3);
        //new Color(0x16, 0xaa, 0xff);

    static String[] playerIcons = {
        "smile25.png",
        "angry25.png",
        "actor25.png",
        "babelfish25.png",
        "bug25.png",
        "cow25.png",               
        //"kitten25.png",
        "butterfly25.png"
    };
    static String[] appleIcons = {
        "apple-red15.png",
        "apple1.png",
        "apple2.png",
        "apple3.png",
        "apple4.png",
        "apple5.png",
        "apple6.png",
    };

    int MAX_PLAYERS = playerIcons.length; 
    int MAX_RES = appleIcons.length;

    public Renderer(RenderingInfo ri) {
        this.ri = ri;
        entitiesRend = new EntitiesRenderer();
        energyRend = new EnergyRenderer();
        
        beings = new Image[MAX_PLAYERS][11];
        beingIcon = new ImageIcon[MAX_PLAYERS];
        
        for (int i=0; i<MAX_PLAYERS; i++) {
            initPlayer(playerIcons[i], beings, i);
            beingIcon[i] = new ImageIcon(beings[i][2]);
        }
        
        resources = new Image[MAX_RES];
        
        for (int i=0; i<MAX_RES; i++) {
            resources[i] = Images.getImage(appleIcons[i]);
        }
    }
    
    private void initPlayer(String name, Image array[][], int idx) {
        for (int i=0; i<11; i++) {
            array[idx][i] = Images.getImage(name).
                getScaledInstance(i*2+10, i*2+10, Image.SCALE_SMOOTH);
            // It looks absurd but the scaled images are loaded lazily,
            // and here's a way to force image preloading
            array[idx][i].getWidth(null);
        }
    }

    class EntitiesRenderer implements Walker<Integer> {
        
        private int energyToSize(float e) {
            if (e < 8f) {
                return 5;
            }
            if (e >= 208f) {
                return 25;
            }
            return (int)((e - 8) / 10f) + 5;
        }
        
        private int sizeToIdx(int size) {
            return (size - 5) / 2;
        }

        public void walk(Integer id) {
	    float energy = ri.getEnergy(id);                
	    Location loc = ri.getLocation(id);
	    int SIZE = energyToSize(energy);
	    int x = getX(loc) + (int)(sx*0.5f)-SIZE/2; 
	    int y = getY(loc) + (int)(sy*0.5f)-SIZE/2;
	    g.drawImage(beings[ri.getOwnerId(id) % MAX_PLAYERS][sizeToIdx(SIZE)], x, y, null);
	    if (Constants.getDrawNumber()) {
		g.setColor(Color.BLACK);
		String ids = id.toString();
		FontMetrics fm = g.getFontMetrics();
		g.drawString(ids, x + SIZE/2-fm.stringWidth(ids)/2, 
			     y + SIZE/2+4);
            }
        }
    }
    
    class EnergyRenderer implements Walker<Location> {
        float minEnergy = 0f, maxEnergy = Constants.MAX_REGULAR;

        private Image resourceFor(float e) {
            if (e < Constants.MAX_REGULAR / 10f) return resources[MAX_RES-1];
            if (e > maxEnergy) return resources[0];

            int i = (int) ((maxEnergy - e) / maxEnergy  * MAX_RES)+1;
            if (i >= MAX_RES) i = MAX_RES-1;
            return resources[i];
        }

        public void walk(Location l) {
            float e = ri.getResourceCount(l);
            int x = getX(l); int y = getY(l);
            g.drawImage(resourceFor(e), x-7, y-5, null);
        }
    }
    
    private void renderPlayfield() {
        final int BORDER = 8;
        
        x0 = BORDER; y0 = BORDER;
        
        int x = w / 2;
        int y = h / 2;
        
        g.setColor(Color.GRAY);
        g.fillRect(0, 0, w, h);
        
        w -= 2 * BORDER;
        h -= 2 * BORDER;
        // compute scaling coeeficients
        java.util.List<Integer> diameters = ri.diameters();        
        if (diameters == null) {
            return;
        }

        int d1 = diameters.get(0), d2 = diameters.get(1);
        sx = (float)w / (float)d1;
        sy = (float)h / (float)d2;
        
        g.setColor(playfieldColor);
        g.fillRect(x0, y0, w, h);
                
        g.setColor(Color.LIGHT_GRAY);
        if (Constants.getDrawGrid()) {
            for (int i=0; i<d1; i++) {
                g.drawLine(getX(i), getY(0), getX(i), getY(d2));
            }
            for (int i=0; i<d2; i++) {
                g.drawLine(getX(0), getY(i), getX(d1), getY(i));
            }
        }
        
    }

    protected void renderResources() {
	ri.forAllR(energyRend);        
    }
    
    protected void renderEntities() {
        ri.forAllE(entitiesRend);
    }

    public void render(int w, int h, Graphics2D g2) {
        // cache in fields
        this.g = g2; this.w = w; this.h = h;
        
        if (ri.getStatus() == RemoteProto.STATUS_UNKNOWN) {
            g.setColor(Color.WHITE);
            g.fillRect(0, 0, w, h);
            g.setColor(Color.RED); 
            g.setFont(new Font("Arial Bold", Font.BOLD, 24));
            g.drawString("Not ready", w/2 - 40, h/2 - 20);
            return;
        }
        
        // in theory renderer should be able to handle various topologies
        // be stupid for now
        renderPlayfield();
        
        // render resources
        renderResources();
        
        // now render every entity in the universe
        renderEntities();

        // cleanup cached values
        this.g = null; this.w = 0; this.h = 0;
        
    }

    public ImageIcon getPlayerIcon(int playerId) {
        return beingIcon[playerId % MAX_PLAYERS];
    }
    
    private int getX(int x) {
        return (int)(sx * x +0.5f) + x0;
    }
    
    private int  getY(int y) {
        return (int)(sy * y +0.5f) + y0;
    }
    
    private int getX(Location loc) {
        return (int)(sx * loc.getX()+0.5f) + x0;
    }
    
    private int  getY(Location loc) {
        return (int)(sy * loc.getY()+0.5f) + y0;
    }
}
